@@ -166,24 +166,12 @@ component_create (const char *name)
 void
 component_free (struct component *comp)
 {
-  size_t i;
-
   component_unlink (comp);
   free (comp->tag);
   free (comp->program);
   free (comp->command);
-  if (comp->argv)
-    {
-      for (i = 0; i < comp->argc; i++)
-	free (comp->argv[i]);
-      free (comp->argv);
-    }
-  if (comp->env)
-    {
-      for (i = 0; comp->env[i]; i++)
-	free (comp->env[i]);
-      free (comp->env);
-    }
+  argv_free (comp->argv);
+  envop_free (comp->envop);
   free (comp->dir);
   grecs_list_free (comp->prereq);
   grecs_list_free (comp->depend);
