@@ -56,6 +56,7 @@
 #include "identity.h"
 #include "acl.h"
 #include "libpies.h"
+#include "envop.h"
 #include "grecs/json.h"
 
 #define ARRAY_SIZE(a) (sizeof(a)/sizeof((a)[0]))
@@ -229,7 +230,7 @@ struct component
   char *command;           /* Full command line */
   size_t argc;             /* Number of command line arguments */
   char **argv;             /* Program command line */
-  char **env;              /* Program environment */
+  envop_t *envop;          /* Environment modification program */
   char *dir;               /* Working directory */
   struct grecs_list *prereq;  /* Prerequisites */
   struct grecs_list *depend;  /* Dependency targets */
@@ -333,6 +334,8 @@ void free_redirector (struct redirector *rp);
 void pies_schedule_action (int act);
 void free_action (struct action *act);
 
+void argv_free (char **argv);
+
 #define PIES_CHLD_NONE             0
 #define PIES_CHLD_CLEANUP          0x01
 #define PIES_CHLD_WAKEUP           0x02
