@@ -49,6 +49,7 @@ char const *instance = "pies";
 char *config_file;   
 char default_config_file[] = SYSCONFDIR "/piesctl.conf";
 int preprocess_only;
+int no_netrc_option;
 int verbose;
 int dump;
 
@@ -543,7 +544,8 @@ shttp_connect (struct pies_url *url, struct grecs_sockaddr *source_addr)
   if (pies_url_copy (&conn->url, url))
     grecs_alloc_die ();
 
-  netrc_scan (conn->url);
+  if (!no_netrc_option)
+    netrc_scan (conn->url);
 
   return conn;
 }
