@@ -1,5 +1,5 @@
 /* This file is part of GNU Pies.
-   Copyright (C) 2013-2016 Sergey Poznyakoff
+   Copyright (C) 2013-2019 Sergey Poznyakoff
 
    GNU Pies is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -1131,7 +1131,7 @@ inittab_parse_line (struct inittab_ctx *ctx)
 	      strerror (errno));
       return inittab_err;
     }
-  wordsplit_getwords (&ws, &comp->argc, &comp->argv);
+  wordsplit_get_words (&ws, &comp->argc, &comp->argv);
   comp->program = grecs_strdup (comp->argv[0]);
   wordsplit_free (&ws);
 
